/*
Author: Cao Thanh Tung
Date: 01/03/2009

File Name: cudaFlipping.cu

This file include all CUDA code to perform the flipping step

===============================================================================

Copyright (c) 2009, School of Computing, National University of Singapore. 
All rights reserved.

Project homepage: http://www.comp.nus.edu.sg/~tants/delaunay.html

If you use GPUDT and you like it or have comments on its usefulness etc., we 
would love to hear from you at <tants@comp.nus.edu.sg>. You may share with us
your experience and any possibilities that we may improve the work/code.

===============================================================================

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the National University of University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission from the National University of Singapore. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#pragma warning(disable: 4311 4312)

#include <device_functions.h>
#include <stdio.h>
#include <string.h>
#include "../gpudt.h"
#include "cuda.h"
#include "common.h"

#include "cudaCCW.h"

/***********************************************************
 * Declarations
 ***********************************************************/
#define WBLOCK      256        

#define SET_TRIANGLE(vOrg, vDest, vApex, nOrg, nDest, nApex, tri, ori) \
    ctriangles[(tri) * 9 + 3 + ((ori) + 1) % 3] = (vOrg); \
    ctriangles[(tri) * 9 + 3 + ((ori) + 2) % 3] = (vDest); \
    ctriangles[(tri) * 9 + 3 + (ori)] = (vApex); \
    ctriangles[(tri) * 9 + 6 + (ori)] = (nOrg); \
    ctriangles[(tri) * 9 + 6 + ((ori) + 1) % 3] = (nDest); \
    ctriangles[(tri) * 9 + 6 + ((ori) + 2) % 3] = (nApex) 

#define UPDATE_TEMP_LINK(pTriOri, pNext) \
    if ((pTriOri) >= 0) \
        ctriangles[decode_tri(pTriOri) * 9 + 6 + decode_ori(pTriOri)] = -(pNext)

#define UPDATE_LINK(pTriOri, pNext) \
    if ((pTriOri) >= 0) \
        ctriangles[decode_tri(pTriOri) * 9 + decode_ori(pTriOri)] = (pNext)

/**************************************************************
 * Exported methods
 **************************************************************/
extern "C" int cudaFlipping(int **suspective); 

/**************************************************************
 * Definitions
 **************************************************************/
// Decode an oriented triangle. 
// An oriented triangle consists of 32 bits. 
// - 30 highest bits represent the triangle index, 
// - 2 lowest bits represent the orientation (the starting vertex, 0, 1 or 2)
#define decode_tri(x)            ((x) >> 2)
#define decode_ori(x)            ((x) & 3)
#define encode_tri(tri, ori)     (((tri) << 2) | (ori))

/************************************************************
 * Variables and functions shared with the main module
 ************************************************************/
extern int nTris, nVerts, nPoints;                
extern int *ctriangles;            
extern REAL2 *covertices;        
extern int *cflag; 

/*************************************************************
 * Detect all sites that can be shifted together 
 * without causing any crossing. 
 * We're guaranteed that all boundary sites are already marked
 *************************************************************/
__global__ void kernelNeedFlipping(int *ctriangles, REAL2 *covertices, BYTE *flipOri, 
                                   int *flipStep, int *flipBy, BYTE *cmarker, 
                                   int nTris, int step, int *cflag) {
    __shared__ int ct[WBLOCK * 9]; 

    int tId = threadIdx.x; 
    int x = (blockIdx.y * gridDim.x + blockIdx.x) * blockDim.x, x9 = x * 9;
    int i, id; 
            
    if (x >= nTris)
        return ;

    // Cooperatively read all triangles processed by one block
    for (i = 0, id = tId; i < 9; i++, id += WBLOCK)
        ct[id] = ctriangles[x9 + id]; 

    __syncthreads(); 
    
    x += tId;

    if (x >= nTris || cmarker[x] != 0) 
        return ; 

    REAL2 pOrg, pDest, pApex, pOpp;
    int nOrg, nDest, nApex; 
    int pTri, pOri; 
    bool notsure = false; 

    nOrg = ct[tId * 9 + 0]; 
    nDest = ct[tId * 9 + 1]; 
    nApex = ct[tId * 9 + 2];

    pOrg = covertices[ct[tId * 9 + 4]]; 
    pDest = covertices[ct[tId * 9 + 5]]; 
    pApex = covertices[ct[tId * 9 + 3]];

    // nOrg
    pTri = decode_tri(nOrg); pOri = decode_ori(nOrg); 
    if (x < pTri || (pTri >= 0 && cmarker[pTri] != 0)) {
        pOpp = covertices[ctriangles[pTri * 9 + 3 + pOri]]; 

        switch (cuda_inCircle(pOrg, pDest, pApex, pOpp)) {
            case 0 : notsure = true; break;     
            case 1 : 
                flipOri[x] = 2; 
                flipStep[x] = step; 
                atomicMin(&flipBy[x], x); 
                atomicMin(&flipBy[pTri], x);
                *cflag = 1; 
                return ; 
        }
    }

    // nDest
    pTri = decode_tri(nDest); pOri = decode_ori(nDest); 
    if (x < pTri || (pTri >= 0 && cmarker[pTri] != 0)) {
        pOpp = covertices[ctriangles[pTri * 9 + 3 + pOri]]; 

        switch (cuda_inCircle(pDest, pApex, pOrg, pOpp)) {
            case 0 : notsure = true; break; 
            case 1 : 
                flipOri[x] = 0; 
                flipStep[x] = step; 
                atomicMin(&flipBy[x], x); 
                atomicMin(&flipBy[pTri], x);
                *cflag = 1; 
                return ; 
        }
    }

    // nApex
    pTri = decode_tri(nApex); pOri = decode_ori(nApex); 
    if (x < pTri || (pTri >= 0 && cmarker[pTri] != 0)) {
        pOpp = covertices[ctriangles[pTri * 9 + 3 + pOri]]; 

        switch (cuda_inCircle(pApex, pOrg, pDest, pOpp)) {
            case 0 : notsure = true; break; 
            case 1 : 
                flipOri[x] = 1; 
                flipStep[x] = step; 
                atomicMin(&flipBy[x], x);       // The one with minimun index
                atomicMin(&flipBy[pTri], x);    // will win. 
                *cflag = 1; 
                return ; 
        }
    }


    if (notsure)    // Inaccurate in_circle test.
        flipStep[x] = -step; 
    else
        cmarker[x] = 1;
}

// To flip a pair of triangle, we need to update the links between neighbor
// triangles. There is a chance that we need to flip two pairs in which 
// there are two neighboring triangles, thus the flipping is performed
// in 3 phases: 
// - Phase 1: Update the triangle vertices. Also, note down the new 
//            neighbors of each triangles in the 3 nexttri links (previously
//            used to store the vertex array). 
// - Phase 2: Each new triangle will then inform its new neighbors of its 
//            existence by writing itself in the corresponding nexttri link
//            of its neighbor.
// - Phase 3: Each new triangle update its links to its neighbors using the
//            3 new nexttri links. If any of the link is not updated, that
//            means the corresponding neighbor is not flipped in this pass. 
//            We then actively update the corresponding link in that triangle.
__global__ void kernelUpdatePhase1(int *ctriangles, BYTE *cflipOri, int *cflipStep, 
                                   int *cflipBy, BYTE *cmarker, int nTris, int step) 
{
    int x = (blockIdx.y * gridDim.x + blockIdx.x) * blockDim.x + threadIdx.x;

    if (x >= nTris)
        return ; 

    if (cflipStep[x] < 0) {  // Inaccurate incircle test
        //cmarker[x] = 1; 
        return ; 
    }

    int ori, pOpp, pOppTri, pOppOri, pOrg, pDest, pApex, nOrg, nApex, npDest, npApex;
    int newtag = -1; 

    if (cflipBy[x] == x) {      // I'm the one who win the right to flip myself
        ori = cflipOri[x]; 
        pOpp = ctriangles[x * 9 + (ori + 1) % 3]; 
        pOppTri = decode_tri(pOpp);
        pOppOri = decode_ori(pOpp); 

        if (cflipBy[pOppTri] == x) {    // I'm also the one who win the right
            newtag = pOpp;              // to flip this neighbor, so I can flip.
            pOrg = ctriangles[x * 9 + 3 + (ori + 1) % 3]; 
            pDest = ctriangles[x * 9 + 3 + (ori + 2) % 3]; 
            pApex = ctriangles[x * 9 + 3 + ori];
            pOpp = ctriangles[pOppTri * 9 + 3 + pOppOri]; 
            nOrg = ctriangles[x * 9 + ori]; 
            nApex = ctriangles[x * 9 + (ori + 2) % 3]; 
            npDest = ctriangles[pOppTri * 9 + (pOppOri + 1) % 3]; 
            npApex = ctriangles[pOppTri * 9 + (pOppOri + 2) % 3]; 

            // Update vertices + nexttri links
            SET_TRIANGLE(pOrg, pDest, pOpp, 3 + nOrg, -1, 3 + nApex, x, ori); 
            SET_TRIANGLE(pApex, pOrg, pOpp, -1, 3 + npDest, 3 + npApex, pOppTri, pOppOri); 
        }
    }

     // Record the opp triangle to be used in the next phase
    cflipStep[x] = newtag;  
}

__global__ void kernelUpdatePhase2(int *ctriangles, BYTE *cflipOri, int *cflipStep, 
                                   int *cflipBy, int nTris, int step) 
{
    int x = (blockIdx.y * gridDim.x + blockIdx.x) * blockDim.x + threadIdx.x;

    if (x >= nTris || cflipStep[x] < 0) 
        return ; 

    int ori, pOpp, pOppTri, pOppOri, nOrg, nApex, npDest, npApex;

    ori = cflipOri[x]; 
    pOpp = cflipStep[x]; 
    pOppTri = decode_tri(pOpp);
    pOppOri = decode_ori(pOpp); 
    nOrg = ctriangles[x * 9 + ori]; 
    nApex = ctriangles[x * 9 + (ori + 2) % 3]; 
    npDest = ctriangles[pOppTri * 9 + (pOppOri + 1) % 3]; 
    npApex = ctriangles[pOppTri * 9 + (pOppOri + 2) % 3]; 

    // Update my neighbors of my existence
    UPDATE_TEMP_LINK(nOrg, encode_tri(x, ori)); 
    UPDATE_TEMP_LINK(nApex, encode_tri(pOppTri, pOppOri)); 
    UPDATE_TEMP_LINK(npDest, encode_tri(x, (ori + 1) % 3)); 
    UPDATE_TEMP_LINK(npApex, encode_tri(pOppTri, (pOppOri + 2) % 3)); 
}

__global__ void kernelUpdatePhase3(int *ctriangles, BYTE *cflipOri, int *cflipStep, 
                                   int *cflipBy, BYTE *cmarker, int nTris, int step) 
{
    int x = (blockIdx.y * gridDim.x + blockIdx.x) * blockDim.x + threadIdx.x;

    if (x >= nTris || cflipStep[x] < 0) 
        return ; 

    int ori, pOpp, pOppTri, pOppOri, nOrg, nApex, npDest, npApex;

    ori = cflipOri[x]; 
    pOpp = cflipStep[x]; 
    pOppTri = decode_tri(pOpp);
    pOppOri = decode_ori(pOpp); 

    // Update other links
    nOrg = ctriangles[x * 9 + 6 + ori]; 
    nApex = ctriangles[x * 9 + 6 + (ori + 2) % 3]; 
    npDest = ctriangles[pOppTri * 9 + 6 + (pOppOri + 1) % 3]; 
    npApex = ctriangles[pOppTri * 9 + 6 + (pOppOri + 2) % 3]; 

    if (nOrg > 0) {        // My neighbor do not update me, update him
        nOrg = -(nOrg - 3); 
        UPDATE_LINK(-nOrg, encode_tri(x, ori)); 
    }

    if (nApex > 0) {
        nApex = -(nApex - 3); 
        UPDATE_LINK(-nApex, encode_tri(pOppTri, pOppOri)); 
    }

    if (npDest > 0) {
        npDest = -(npDest - 3); 
        UPDATE_LINK(-npDest, encode_tri(x, (ori + 1) % 3)); 
    }

    if (npApex > 0) {
        npApex = -(npApex - 3); 
        UPDATE_LINK(-npApex, encode_tri(pOppTri, (pOppOri + 2) % 3)); 
    }

    // Update my own links
    ctriangles[x * 9 + ori] = -nOrg; 
    ctriangles[x * 9 + (ori + 1) % 3] = -npDest; 
    ctriangles[x * 9 + (ori + 2) % 3] = encode_tri(pOppTri, (pOppOri + 1) % 3); 
    ctriangles[pOppTri * 9 + pOppOri] = -nApex; 
    ctriangles[pOppTri * 9 + (pOppOri + 1) % 3] = encode_tri(x, (ori + 2) % 3); 
    ctriangles[pOppTri * 9 + (pOppOri + 2) % 3] = -npApex; 

    // Mark the affected triangles, so that we will check again in the next pass
    cflipStep[x] = -1; 
    if (nOrg <= 0) cmarker[decode_tri(-nOrg)] = 0; 
    if (nApex <= 0) cmarker[decode_tri(-nApex)] = 0; 
    if (npDest <= 0) cmarker[decode_tri(-npDest)] = 0; 
    if (npApex <= 0) cmarker[decode_tri(-npApex)] = 0; 
}

//////////////////////////////////////////////////////////////

int cudaFlipping(int **suspective) 
{
    int flag; 
    int step = 1; 

    // Constants for the EXACT tests
    cutilSafeCall( cudaMemcpyToSymbol(constData, hostConst, 13 * sizeof(REAL)) ); 

    int *cflipStep, *cflipBy;
    BYTE *cmarker, *cflipOri; 
    dim3 grid, block; 

    cflipOri = (BYTE *) cudaMemAlloc(2); 
    cflipStep = cudaMemAlloc(2 * sizeof(int)); 
    cflipBy = cudaMemAlloc(2 * sizeof(int)); 
    cmarker = (BYTE *) cudaMemAlloc(2); 

    cutilSafeCall( cudaMemset(cflipStep, 255, nTris * sizeof(int)) ); 
    cutilSafeCall( cudaMemset(cmarker, 0, nTris) ); 

    // Start flipping
    do {
        cutilSafeCall( cudaMemset(cflag, 0, sizeof(int)) ); 
        cutilSafeCall( cudaMemset(cflipBy, 127, nTris * sizeof(int)) ); 

        // Perform incircle tests
        block = dim3(WBLOCK); 
        grid = dim3(STRIPE, nTris / (STRIPE * block.x) + 1); 
        kernelNeedFlipping<<< grid, block >>>(ctriangles, covertices, cflipOri, cflipStep, 
            cflipBy, cmarker, nTris, step, cflag); 
        cutilCheckError(); 

        cutilSafeCall( cudaMemcpy(&flag, cflag, sizeof(int), cudaMemcpyDeviceToHost) ); 

        if (flag > 0) {
            // DEBUG
            //int *flipStep = (int *) malloc(nTris * sizeof(int)); 
            //cudaMemcpy(flipStep, cflipStep, nTris * sizeof(int), cudaMemcpyDeviceToHost); 

            //int count = 0; 

            //for (int i = 0; i < nTris; i++)
            //    if (flipStep[i] >= 0)
            //        count++; 

            //free(flipStep);

            //printf("     Step %i: Possible flip %i edges\n", step, count); 
            //totalCheck += count; 
            // DEBUG

            kernelUpdatePhase1<<< grid, block >>>(ctriangles, cflipOri, cflipStep, 
                cflipBy, cmarker, nTris, step); 

            // DEBUG
            //flipStep = (int *) malloc(nTris * sizeof(int)); 
            //cudaMemcpy(flipStep, cflipStep, nTris * sizeof(int), cudaMemcpyDeviceToHost); 

            //count = 0; 

            //for (int i = 0; i < nTris; i++)
            //    if (flipStep[i] >= 0) {
            //        //printf("     Can flip %i and %i\n", i, flipStep[i] >> 2); 
            //        count++; 
            //    }

            //free(flipStep);

            //printf("     Step %i: Can flip %i edges\n", step, count); 
            //totalFlip += count; 
            // DEBUG

            kernelUpdatePhase2<<< grid, block >>>(ctriangles, cflipOri, cflipStep, 
                cflipBy, nTris, step); 

            kernelUpdatePhase3<<< grid, block >>>(ctriangles, cflipOri, cflipStep, 
                cflipBy, cmarker, nTris, step); 
        }

        step++; 
    } while (flag > 0); 

    // All the suspectice incircle tests will be noted down 
    // and let the CPU perform the exact incircle tests.
    int marker = step - 1; 

    *suspective = (int *) malloc(nTris * sizeof(int)); 
    cutilSafeCall( cudaMemcpy(*suspective, cflipStep, nTris * sizeof(int), cudaMemcpyDeviceToHost) ); 

    //printf("     Flipping - Total step: %i ; Total time: %.5fms\n", step-1, totalTime); 
    //printf("     Total flip = %i / %i\n", totalFlip, totalCheck); 
 
    /*******************************
     * DONE
     *******************************/

    cudaMemFree(cmarker); 
    cudaMemFree(cflipBy); 
    cudaMemFree(cflipStep); 
    cudaMemFree(cflipOri);  

    return marker; 
}