/*
Author: Cao Thanh Tung, Rong Guodong & Stephanus
Date: 26/01/2010

File Name: main.cpp

Example of the usage of GPU-DT (WITH VISUALIZATION)

===============================================================================

Copyright (c) 2009, 2010, School of Computing, National University of Singapore. 
All rights reserved.

Project homepage: http://www.comp.nus.edu.sg/~tants/delaunay.html

If you use GPUDT and you like it or have comments on its usefulness etc., we 
would love to hear from you at <tants@comp.nus.edu.sg>. You may share with us
your experience and any possibilities that we may improve the work/code.

===============================================================================

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the National University of University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission from the National University of Singapore. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/


#ifdef _WIN32
#  define WINDOWS_LEAN_AND_MEAN
#  include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>

#include <GL/glut.h>
#include <gpudt/gpudt.h>

//**********************************************************************************************
//* Testing
//**********************************************************************************************
int nVertices   = 100;

// Global Vars
int winWidth  = 512;
int winHeight = 512;

int fboSize  = 512;    // Min is 256

// Mouse Control
int oldMouseX,oldMouseY;
int currentMouseX,currentMouseY;
float clickedX = 0.0f, clickedY = 0.0f; 
bool isLeftMouseActive;
bool isRightMouseActive;

// States
float zDepth          = 1.0f;
float xTranslate      = 0.0f;
float yTranslate      = 0.0f;

// GPUDT Vars
GPUDTPARAMS  gpudtparams;
PGPUDTOUTPUT pOutput = NULL;
gpudtTriangle *clickedTri = NULL; 

//**********************************************************************************************
//* Random Point Generator
//**********************************************************************************************

// Random number generator, obtained from http://oldmill.uchicago.edu/~wilder/Code/random/
unsigned long z, w, jsr, jcong; // Seeds
void randinit(unsigned long x_) 
{ z =x_; w = x_; jsr = x_; jcong = x_; }
unsigned long znew() 
{ return (z = 36969 * (z & 0xfffful) + (z >> 16)); }
unsigned long wnew() 
{ return (w = 18000 * (w & 0xfffful) + (w >> 16)); }
unsigned long MWC()  
{ return ((znew() << 16) + wnew()); }
unsigned long SHR3()
{ jsr ^= (jsr << 17); jsr ^= (jsr >> 13); return (jsr ^= (jsr << 5)); }
unsigned long CONG() 
{ return (jcong = 69069 * jcong + 1234567); }
unsigned long rand_int()         // [0,2^32-1]
{ return ((MWC() ^ CONG()) + SHR3()); }
REAL random()     // [0,1]
{ return ((REAL) rand_int() / (REAL(ULONG_MAX)+1)); }

// Generate a randomset of points (Continuous Space)
void GenerateRandomPoints(PGPUDTPARAMS pParams, int nPoints)
{    
    REAL x, y;
    randinit(0);

	pParams->nPoints = nPoints;
    pParams->points  = new REAL[nPoints * 2];

    int idx = 0;

    for (int i=0; i<nPoints; ++i)
    {
        x = random();
        y = random();

		pParams->points[idx++] = pParams->minX + (pParams->maxX - pParams->minX) * x;    // x
        pParams->points[idx++] = pParams->minY + (pParams->maxY - pParams->minY) * y;    // y    
    }    

}

void ReleaseRandomPoints(PGPUDTPARAMS pParams)
{
    pParams->nPoints = 0;
    if (pParams->points) delete [] pParams->points;
}

//**********************************************************************************************
//* Init / Exit
//**********************************************************************************************
void exitGL(int returnCode)
{
    // Release Input
    ReleaseRandomPoints(&gpudtparams);

	// Release GPUDT
	gpudtReleaseDTOutput(pOutput); 

    exit(returnCode);
}

//**********************************************************************************************
//* GLUT Callback
//**********************************************************************************************
void glutReshape(int width,int height)
{
    winWidth  = width;
    winHeight = height;
}

void drawTri(gpudtTriangle *tri) {
    if (tri == NULL)
        return ; 

    int id0 = tri->vtx[0]; 
    int id1 = tri->vtx[1]; 
    int id2 = tri->vtx[2];

    glVertex2d(pOutput->vertices[id0].x, pOutput->vertices[id0].y);    
    glVertex2d(pOutput->vertices[id1].x, pOutput->vertices[id1].y);    
    glVertex2d(pOutput->vertices[id2].x, pOutput->vertices[id2].y);    
}

void drawTriangles()
{
    gpudtOrientedTriangle otri; 

    // Draw the selected triangle
    if (clickedTri != NULL) {
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	    glBegin(GL_TRIANGLES);
            glColor3f(0.0, 0.0, 1.0);
            drawTri(clickedTri); 
            glColor3f(0.0, 1.0, 0.0);
            decode(clickedTri->tri[0], otri); 
            drawTri(otri.tri); 
            decode(clickedTri->tri[1], otri); 
            drawTri(otri.tri); 
            decode(clickedTri->tri[2], otri); 
            drawTri(otri.tri); 
        glEnd();
    }

    // Draw Triangles
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glColor3f(1.0, 1.0, 1.0);

	glBegin(GL_TRIANGLES);
	    for (int i=0; i<pOutput->nTris; ++i) 
            drawTri(&pOutput->triangles[i]); 
    glEnd();

    // Draw the scaled points
    glPointSize(3.0);  
	glBegin(GL_POINTS);
    if (clickedTri == NULL) {
        glColor3f(0.0, 0.0, 1.0); 
        glVertex2d(clickedX, clickedY); 
    }
    glColor3f(1.0, 0.0, 0.0);
    for (int i=0; i<gpudtparams.nPoints; ++i)
        glVertex2d(pOutput->vertices[i].x, pOutput->vertices[i].y);
    glEnd();
	glPointSize(1.0);   
}

void glutDisplay()
{
    //******************************************************************************************
    // Visualization
    //******************************************************************************************
    glPushAttrib(GL_ALL_ATTRIB_BITS);
        glClearColor(0.0, 0.0, 0.0, 0.0);
        glClear(GL_COLOR_BUFFER_BIT);

        // Setup Matrices
        glViewport(0,0,winWidth,winHeight);
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        gluOrtho2D(gpudtparams.minX, gpudtparams.maxX, gpudtparams.minY, gpudtparams.maxY);    

        REAL centerX = (gpudtparams.minX + gpudtparams.maxX) / 2.0;
        REAL centerY = (gpudtparams.minY + gpudtparams.maxY) / 2.0;
        
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        glTranslatef(xTranslate*zDepth, yTranslate*zDepth, 0.0);
        glScalef(zDepth, zDepth, 1.0);

        // Draw Triangles
        drawTriangles();
    glPopAttrib();

    glutSwapBuffers();
}

gpudtTriangle *locateTriangle(PGPUDTOUTPUT pOutput, REAL x, REAL y)
{
	gpudtOrientedTriangle olooptri, oFirstTri, temptri;
	gpudtTriangle *ptr;
	int triOrg, triDest, triApex, vcenter;
	gpudtVertex v;
    REAL ccTest, ccTest1, ccTest2;
	bool bONext;

    // Search for this vertex
    v.x = x; v.y = y; 

    // Start from an arbitrary triangle
    olooptri.orient = 0; 
    olooptri.tri = &pOutput->triangles[0]; 
    org(olooptri, vcenter); 

	//**********************************************************************************************
	// Record the first starting triangle
	//**********************************************************************************************
	otricopy(olooptri, oFirstTri);

	org(olooptri, triOrg);
	dest(olooptri, triDest);
	apex(olooptri, triApex);
	ccTest = counterclockwise(&pOutput->vertices[triOrg], &pOutput->vertices[triDest], &v);
	bONext = (ccTest >= 0.0);

    // Walk around the triangle fan
	while (true)
	{
		ccTest = bONext ? 
			counterclockwise(&pOutput->vertices[triOrg], &v, &pOutput->vertices[triApex]) :
			counterclockwise(&pOutput->vertices[triOrg], &pOutput->vertices[triDest], &v);

		if (ccTest >= 0.0)  // Inside this angle of the triangle fan
		{
			if (counterclockwise(&pOutput->vertices[triDest], &pOutput->vertices[triApex], &v) >= 0.0) // Found
                return olooptri.tri; 

            // Start walking to find the triangle
			lnextself(olooptri);
			sym(olooptri, temptri);
			while (true)
			{
				if (temptri.tri == NULL) // Outside the mesh, triangle not found
                    return NULL; 

				lprev(temptri, olooptri);
				org(olooptri, triOrg);
				dest(olooptri, triDest);
				apex(olooptri, triApex);

				ccTest1 = counterclockwise(&pOutput->vertices[triOrg], &pOutput->vertices[triDest], &v);
				ccTest2 = counterclockwise(&pOutput->vertices[triApex], &pOutput->vertices[triOrg], &v);

				if (ccTest1 >= 0 && ccTest2 >= 0) // Found
                    return olooptri.tri; 
				else
				{
                    // Which way should we continue the search?
					REAL ccTestCenter = counterclockwise(&pOutput->vertices[triOrg], &pOutput->vertices[vcenter], &v);
					if (ccTestCenter>0)
					{
						lprevself(olooptri);
						sym(olooptri, temptri);
					}
					else //if (ccTestCenter<0)
						sym(olooptri, temptri);
				}
			}
		}

		// goto the next triangle
		otricopy(olooptri, temptri);
		if (bONext)
		{			
			onextself(olooptri);

			if (olooptri.tri == NULL) // Outside the mesh, not found
                return NULL; 
		}
		else
		{
			oprevself(olooptri);

			if (olooptri.tri == NULL) // Outside the mesh, not found
                return NULL; 
		}

		if (otriequal(olooptri, oFirstTri)) // One round, not found
			return NULL; 

		org(olooptri, triOrg);
		dest(olooptri, triDest);
		apex(olooptri, triApex);
	}

    return NULL; 
}

void glutMouse(int button,int state,int x,int y)
{
    int modifiers = glutGetModifiers(); 
    
    if (state == GLUT_UP)
        switch (button)
    {
        case GLUT_LEFT_BUTTON:
            isLeftMouseActive = false;
            break;
        case GLUT_RIGHT_BUTTON:
            isRightMouseActive = false;
            break;
    }

    if (state == GLUT_DOWN)
    {
        oldMouseX = currentMouseX = x;
        oldMouseY = currentMouseY = y;

        switch (button)
        {
        case GLUT_LEFT_BUTTON:
            if ((modifiers & GLUT_ACTIVE_SHIFT) > 0) {
                //glTranslatef(xTranslate*zDepth, yTranslate*zDepth, 0.0);
                //glScalef(zDepth, zDepth, 1.0);
                REAL realX = double(x) / winWidth * (gpudtparams.maxX - gpudtparams.minX) + gpudtparams.minX; 
                REAL realY = double(winHeight - y) / winHeight * (gpudtparams.maxY - gpudtparams.minY) + gpudtparams.minY; 
                clickedX = realX / zDepth - xTranslate; 
                clickedY = realY / zDepth - yTranslate; 
                clickedTri = locateTriangle(pOutput, clickedX, clickedY);                    
                glutPostRedisplay(); 
            }            
            else
                isLeftMouseActive = true;
                        
            break;
        case GLUT_RIGHT_BUTTON:
            isRightMouseActive = true;
            break;
        }

    }
}

void glutMouseMotion(int x,int y)
{
    currentMouseX = x;
    currentMouseY = y;

    if (isLeftMouseActive)
    {
        xTranslate += (currentMouseX - oldMouseX) / (zDepth / 2);
        yTranslate -= (currentMouseY - oldMouseY) / (zDepth / 2);
        glutPostRedisplay();
    }
    else if (isRightMouseActive)
    {
        zDepth -= (currentMouseY - oldMouseY) * zDepth / 400.0;
        glutPostRedisplay();
    }

    oldMouseX = currentMouseX;
    oldMouseY = currentMouseY;
}

int main(int argc,char **argv)
{
    glutInit(&argc,argv);
    glutInitWindowPosition(0, 0);
    glutInitWindowSize(winWidth, winHeight);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
    glutCreateWindow("GPU-DT");

    glutDisplayFunc(glutDisplay);
    glutReshapeFunc(glutReshape);
    glutMouseFunc(glutMouse);
    glutMotionFunc(glutMouseMotion);
    
	// Initialize GPUDT
    gpudtparams.fboSize         = fboSize;
    gpudtparams.minX            = 0;
    gpudtparams.minY            = 0;
    gpudtparams.maxX            = fboSize - 3;
    gpudtparams.maxY            = fboSize - 3;
    gpudtparams.points          = NULL;
    gpudtparams.nPoints         = nVertices;

	printf("GPUDT - A Delaunay Triangulation generator using the GPU\n"); 
	printf("\n"); 

	printf("Generating input points...\n"); 
    GenerateRandomPoints(&gpudtparams, nVertices);

	// Run GPUDT
	printf("Running GPUDT..."); 
    pOutput = gpudtComputeDT(&gpudtparams);
	printf(" Done!\n"); 

	printf("Total vertices : %i\n", nVertices); 
	printf("Total triangles: %i\n", pOutput->nTris); 

	// Visualization
	printf("\n"); 
	printf("Visualization\n"); 
	printf("  - Left  mouse + Motion : Move\n"); 
	printf("  - Right mouse + Motion : Zoom\n"); 
    printf("  - Shift + Left mouse   : Highlight a triangle and its 3 neighbors\n"); 

	glutMainLoop();

	exitGL(0);
    return 0;
}